import numpy as np
from math import sqrt, sin, cos, acos, atan2, trunc, pi
import copy

from phoebe import u
from phoebe import c
import libphoebe

import logging
logger = logging.getLogger("MESH")
logger.addHandler(logging.NullHandler())

def compute_volume(sizes, centers, normals):
    """
    Compute the numerical volume of a convex mesh

    :parameter array sizes: array of sizes of triangles
    :parameter array centers: array of centers of triangles (x,y,z)
    :parameter array normals: array of normals of triangles (will normalize if not already)
    :return: the volume (float)
    """
    # the volume of a slanted triangular cone is A_triangle * (r_vec dot norm_vec) / 3.

    # TODO: implement normalizing normals into meshing routines (or at least have them supply normal_mags to the mesh)


    # TODO: remove this function - should now be returned by the meshing algorithm itself
    # although wd method may currently use this
    normal_mags = np.linalg.norm(normals, axis=1) #np.sqrt((normals**2).sum(axis=1))
    return np.sum(sizes*((centers*normals).sum(axis=1)/normal_mags)/3)


def euler_trans_matrix(etheta, elongan, eincl):
    """
    Get the transformation matrix R to translate/rotate a mesh according to
    euler angles.

    The matrix is

      R(long,incl,theta) =
        Rz(pi).Rz(long).Rx(incl).Rz(theta)
        Rz(long).Rx(-incl).Rz(theta).Rz(pi)

    where

      Rx(u) = 1,  0,        0
              0,  cos(u),   -sin(u)
              0,  sin(u),   cos(u)

      Ry(u) = cos(u),   0,  sin(u)
              0,        1,  0
              -sin(u),  0,  cos(u)

      Rz(u) = cos(u),   -sin(u),  0
              sin(u),   cos(u),   0
              0,        0,        1

    Rz(pi) = reflection across z-axis

    Note:

      R(0,0,0) = -1,  0,  0
                  0,  -1, 0
                  0,  0,  1

    :parameter float etheta: euler theta angle
    :parameter float elongan: euler long of asc node angle
    :parameter float eincl: euler inclination angle
    :return: matrix with size 3x3
    """

    s1 = sin(eincl);
    c1 = cos(eincl);
    s2 = sin(elongan);
    c2 = cos(elongan);
    s3 = sin(etheta);
    c3 = cos(etheta);
    c1s3 = c1*s3;
    c1c3 = c1*c3;

    return np.array([
                        [-c2*c3+s2*c1s3, c2*s3+s2*c1c3, -s2*s1],
                        [-s2*c3-c2*c1s3, s2*s3-c2*c1c3, c2*s1],
                        [s1*s3, s1*c3, c1]
                    ])

def Rx(x):
  c = cos(x)
  s = sin(x)
  return np.array([[1., 0., 0.], [0., c, -s], [0., s, c]])

def Ry(x):
  c = cos(x)
  s = sin(x)
  return np.array([[c, 0., -s], [0., 1., 0.], [s, 0., c]])

def Rz(x):
  c = cos(x)
  s = sin(x)
  return np.array([[c, -s, 0.], [s, c, 0.], [0., 0., 1.]])


def spin_in_system(incl, long_an):
    """
    Spin in the plane of sky of a star given its inclination and "long_an"

       incl - inclination of the star in the plane of sky
       long_an - longitude of ascending node (equator) of the star in the plane of sky

    Return:
        spin - in plane of sky
    """
    #print "*** spin_in_system", incl, long_an, np.dot(Rz(long_an), np.dot(Rx(-incl), np.array([0,0,1])))
    # Rz(long_an) Rx(incl) [0, 0, 1]
    return np.dot(Rz(long_an), np.dot(Rx(-incl), np.array([0.,0.,1.])))

def spin_in_roche(s, etheta, elongan, eincl):
  """
    Transform the spin s of a star on Kerpler orbit with

      etheta  - true anomaly
      elongan - longitude of ascending node
      eincl - inclination

    from in the plane of sky reference frame into
    the Roche reference frame.
  """
  #  m = Rz(long).Rx(-incl).Rz(theta).Rz(pi)
  m = euler_trans_matrix(etheta, elongan, eincl)

  return np.dot(m.T, s)


def general_rotation_matrix(theta, phi, alpha):
  """
  Rotation around vector
    u = (sin(theta) cos(phi), sin(theta) sin(phi), cos(theta))
  by an angle
    alpha
  Ref:
    http://ksuweb.kennesaw.edu/~plaval//math4490/rotgen.pdf

  :parameter float theta:
  :parameter float phi:
  :parameter float alpha: rotation angle

  :return: 3x3 matrix of floats
  """

  C = cos(alpha)
  S = sin(alpha)
  t  = 1 - C

  ux = sin(theta)*cos(phi)
  uy = sin(theta)*sin(phi)
  uz = cos(theta)

  return np.array([
                    [t*ux**2 + C, t*ux*uy - S*uz, t*ux*uz + S*uy],
                    [t*ux*uy + S*uz, t*uy**2 + C, t*uy*uz - S*ux],
                    [t*ux*uz - S*uy, t*uy*uz + S*ux, t*uz**2 + C]
                  ])


def transform_position_array(array, pos, euler, is_normal, reverse=False):
    """
    Transform any Nx3 position array by translating to a center-of-mass 'pos'
    and applying an euler transformation

    :parameter array array: numpy array of Nx3 positions in the original (star)
        coordinate frame
    :parameter array pos: numpy array with length 3 giving cartesian
        coordinates to offset all positions
    :parameter array euler: euler angles (etheta, elongan, eincl) in radians
    :parameter bool is_normal: whether each entry is a normal vector rather
        than position vector.  If true, the quantities won't be offset by
        'pos'
    :return: new positions array with same shape as 'array'.
    """
    trans_matrix = euler_trans_matrix(*euler)

    if not reverse:
        trans_matrix = trans_matrix.T

    if isinstance(array, ComputedColumn):
        array = array.for_computations

    if is_normal:
        # then we don't do an offset by the position
        return np.dot(np.asarray(array), trans_matrix)
    else:
        return np.dot(np.asarray(array), trans_matrix) + np.asarray(pos)

def transform_velocity_array(array, pos_array, vel, euler, rotation_vel=(0,0,0)):
    """
    Transform any Nx3 velocity vector array by adding the center-of-mass 'vel',
    accounting for solid-body rotation, and applying an euler transformation.

    :parameter array array: numpy array of Nx3 velocity vectors in the original
        (star) coordinate frame
    :parameter array pos_array: positions of the elements with respect to the
        original (star) coordinate frame.  Must be the same shape as 'array'.
    :parameter array vel: numpy array with length 3 giving cartesian velocity
        offsets in the new (system) coordinate frame
    :parameter array euler: euler angles (etheta, elongan, eincl) in radians
    :parameter array rotation_vel: vector of the rotation velocity of the star
        in the original (star) coordinate frame
    :return: new velocity array with same shape as 'array'
    """

    trans_matrix = euler_trans_matrix(*euler)

    # v_{rot,i} = omega x r_i    with  omega = rotation_vel
    rotation_component = np.cross(rotation_vel, pos_array, axisb=1)
    orbital_component = np.asarray(vel)

    if isinstance(array, ComputedColumn):
        array = array.for_computations

    new_vel = np.dot(np.asarray(array)+rotation_component, trans_matrix.T) + orbital_component

    return new_vel


def wd_grid_to_mesh_dict(the_grid, q, F, d):
    """
    Transform a wd-style mesh to the format used by PHOEBE. Namely this handles
    translating vertices from Nx9 to Nx3x3 and creating the array of indices
    for each triangle.

    :parameter record-array the_grid: output from discretize_wd_style
    :parameter float q: mass-ratio (M_this/M_sibling)
    :parameter float F: syncpar
    :parameter float d: instantaneous unitless separation
    :return: the dictionary in PHOEBE's format to be passed to a Mesh class
    """

    # WD returns a list of triangles with 9 coordinates (v1x, v1y, v1z, v2x, ...)
    triangles_9N = the_grid[:,4:13]

    new_mesh = {}
    # force the mesh to be computed at centers rather than the PHOEBE default
    # of computing at vertices and averaging for the centers.  This will
    # propogate to all ComputedColumns, which means we'll fill those quanities
    # (ie normgrads, velocities) per-triangle.
    new_mesh['compute_at_vertices'] = False
    # PHOEBE's mesh structure stores vertices in an Nx3 array
    new_mesh['vertices'] = triangles_9N.reshape(-1,3)
    # and triangles as indices pointing to each of the 3 vertices (Nx3)
    new_mesh['triangles'] = np.arange(len(triangles_9N)*3).reshape(-1,3)
    new_mesh['centers'] = the_grid[:,0:3]

    new_mesh['tnormals'] = the_grid[:,13:16]
    norms = np.linalg.norm(new_mesh['tnormals'], axis=1)
    new_mesh['normgrads'] = norms
    # TODO: do this the right way by dividing along axis=1 (or using np.newaxis as done for multiplying in ComputedColumns)
    new_mesh['tnormals'] = np.array([tn/n for tn,n in zip(new_mesh['tnormals'], norms)])

    # NOTE: there are no vnormals in wd-style mesh

    new_mesh['areas'] = the_grid[:,3]
    new_mesh['tareas'] = the_grid[:,18]

    # TESTING ONLY - remove this eventually ??? (currently being used
    # to test WD-style eclipse detection by using theta and phi (lat and long)
    # to determine which triangles are in the same "strip")
    new_mesh['thetas'] = the_grid[:,16]
    new_mesh['phis'] = the_grid[:,17]

    # TODO: get rid of this list comprehension
    # grads = np.array([libphoebe.roche_gradOmega_only(q, F, d, c) for c in new_mesh['centers']])

    # new_mesh['normgrads'] = np.sqrt(grads[:,0]**2+grads[:,1]**2+grads[:,2]**2)
    # new_mesh['normgrads'] = norms #np.linalg.norm(grads, axis=1)

    # TODO: actually compute the numerical volume (find old code)
    new_mesh['volume'] = compute_volume(new_mesh['areas'], new_mesh['centers'], new_mesh['tnormals'])
    # new_mesh['area'] # TODO: compute surface area??? (not sure if needed)
    new_mesh['velocities'] = np.zeros(new_mesh['centers'].shape)

    return new_mesh


class ComputedColumn(object):
    """
    Any non-geometric column in a Mesh should be a ComputedColumn.  Depending
    on the type of mesh (defined by mesh._compute_at_vertices), the computed
    column will either compute values at the vertices and return the weighted
    averages across a triangle when observing, OR compute at the centers and
    return that single value when observing.

    This class simply provides a single interface with logical/intuitive
    (hopefully) properties for setting and retrieving values for either
    type of mesh.
    """
    def __init__(self, mesh, value_for_computations=None, **kwargs):
        """
        TODO: add documentation
        """

        self._mesh = mesh

        # NOTE: it is ESSENTIAL that all of the following are np.array
        # and not lists... but is up to the user (phoebe backend since the
        # user will probably not dig this deep)
        self._vertices = None
        self._centers = None

        self._compute_at_vertices = kwargs.get('compute_at_vertices', self.mesh._compute_at_vertices)

        if value_for_computations is not None:
            self.set_for_computations(value_for_computations)


    def __len__(self):
        if self.for_computations is None:
            # this is a bit of a hack, but allows us to check something like
            # if column is None or not len(column) without needing extra
            # conditions for computed columns
            return 0
        return len(self.for_computations)

    @property
    def shape(self):
        return self.for_computations.shape

    def __getitem__(self, k):
        return self.for_computations[k]

    @property
    def compute_at_vertices(self):
        return self._compute_at_vertices

    @property
    def mesh(self):
        """
        Access to the parent mesh

        :return: an instantiated :class:`Mesh`, :class:`ScaledProtoMesh`, or
            :class:`ProtoMesh` object.
        """
        return self._mesh

    @property
    def vertices(self):
        """
        Access to the quantities defined at the vertices, if applicable

        :return: numpy array
        """
        return self._vertices

    @property
    def vertices_per_triangle(self):
        """
        Access to the quantities defined at the vertices, but reshaped to be
        in the order of triangles, with three entries each

        :return: numpy array
        """
        if self.vertices is not None:
            return self.vertices[self.mesh.triangles]
        else:
            return None

    @property
    def centers(self):
        """
        Access to the quantities at the centers of each triangles.  In the
        case where the quantities are provided at the vertices, this will
        return the average of those values.

        :return: numpy array
        """
        if self.mesh._compute_at_vertices:
            return self.averages
        else:
            return self._centers

    @property
    def for_computations(self):
        """
        Access to the quantities as they should be used for computations
        (either .vertices or .centers depending on the value of
        mesh._compute_at_vertices)

        :return: numpy array
        """
        if self.mesh._compute_at_vertices:
            return self.vertices
        else:
            return self.centers

    @property
    def averages(self):
        """
        Access to the average of the values at the vertices for each triangle.
        If the quantities are defined at centers instead of vertices, this
        will return None.  Also see :method:`centers`.

        :return: numpy array or None
        """
        if not self.mesh._compute_at_vertices:
            return None

        return np.mean(self.vertices_per_triangle, axis=1)

    @property
    def weighted_averages(self):
        """
        Access to the weighted averages of the values at the vertices for each
        triangle based on the weights provided by mesh.weights.  This is most
        useful for partially visible triangles when using libphoebe's
        eclipse detection that returns weights for each vertex.

        Note that weights by default are set to 1/3 for each vertex, meaning
        that this will provide the same values as :meth:`averages` unless
        the weights are overridden within the mesh.

        If the quantities are defined at centers instead of vertices, this will
        return None.

        :return: numpy array or None
        """
        if not self.mesh._compute_at_vertices:
            return None

        vertices_per_triangle = self.vertices_per_triangle
        if vertices_per_triangle.ndim==2:
            # return np.dot(self.vertices_per_triangle, self.mesh.weights)
            return np.sum(vertices_per_triangle*self.mesh.weights, axis=1)
        elif vertices_per_triangle.ndim==3:
            return np.sum(vertices_per_triangle*self.mesh.weights[:,np.newaxis], axis=1)
        else:
            raise NotImplementedError

    @property
    def for_observations(self):
        """
        Access to the quantities as they should be used for observations.
        When defined at centers, this will return :meth:`centers`, when defined
        at vertices, this will return :meth:`weighted_averages`.

        :return: numpy array
        """
        if self.mesh._compute_at_vertices:
            # TODO: make this an option at some point?
            # return self.averages
            # for now this can just be achieved by calling self.centers directly
            return self.weighted_averages
        else:
            return self.centers

    def set_for_computations(self, value):
        """
        Set the quantities, either at the vertices or centers depending on the
        settings of the mesh (mesh._compute_at_vertices)
        """
        if self.mesh._compute_at_vertices:
            self._vertices = value
        else:
            self._centers = value


class ProtoMesh(object):
    """
    ProtoMesh is in units of d with the origin at the COM of the STAR.

    Because there is no orbital, scale, or orientation information, those
    fields are not available until the mesh is scaled and/or placed in
    orbit (by using the class constructors on ScaledProtoMesh or Mesh
    respectively).
    """

    def __init__(self, compute_at_vertices=True, **kwargs):
        """
        TODO: add documentation

        :parameter bool compute_at_vertices: whether all
            :class:`ComputedColumn` (physical quantities) should be computed
            at the vertices and then averaged or computed at the centers.
            Generally this should be decided based on whether the vertices
            or centers of the elements are placed on the surface of the
            equipotential.
        """

        self._compute_at_vertices = compute_at_vertices

        self._pvertices         = None  # Vx3
        self._vertices          = None  # Vx3

        self._triangles         = None  # Nx3
        self._centers           = None  # Nx3
        self._areas             = None  # Nx1
        self._tareas            = None  # Nx1

        self._velocities        = ComputedColumn(mesh=self)  # Nx1

        # v = vertices
        # t = triangle
        self._vnormals          = None  # Vx3
        self._tnormals          = None  # Nx3

        self._normgrads         = ComputedColumn(mesh=self)

        self._volume            = None  # scalar
        self._area              = None  # scalar

        ### TESTING FOR WD METHOD ###
        self._phis               = None # Nx1
        self._thetas             = None # Nx1

        ### PHYSICAL QUANTITIES
        self._loggs             = ComputedColumn(mesh=self)
        self._gravs             = ComputedColumn(mesh=self)
        self._teffs             = ComputedColumn(mesh=self)
        self._abuns             = ComputedColumn(mesh=self)
        self._irrad_frac_refl  = ComputedColumn(mesh=self)
        # self._frac_heats          = ComputedColumn(mesh=self)
        # self._frac_scatts          = ComputedColumn(mesh=self)


        self._pos               = np.array([0.,0.,0.])  # will be updated when placed in orbit (only for Meshes)
        self._pos_center        = np.array([0.,0.,0.])  # will be updated when placed in orbit (only for Meshes)
        self._scalar_fields     = ['volume', 'area']

        # TODO: split keys that are set vs computed-on-the-fly so when
        # we call something like ScaledProtoMesh.from_proto we don't have
        # to do all the on-the-fly computations just to discard them because
        # they aren't setable.
        keys = ['compute_at_vertices',
                  'pvertices', 'vertices', 'triangles', 'centers',
                  'coords_for_computations', 'normals_for_computations',
                  'rs', 'rprojs', 'cosbetas',
                  'areas', 'tareas', 'areas_si',
                  'velocities', 'vnormals', 'tnormals',
                  'normgrads', 'volume', 'area',
                  'phis', 'thetas',
                  'loggs', 'gravs', 'teffs', 'abuns', 'irrad_frac_refl'] # frac_heats, frac_scatts
        self._keys = keys + kwargs.pop('keys', [])

        self.update_columns(**kwargs)

    def __getitem__(self, key):
        """
        TODO: add documentation
        """

        # TODO: split this stuff into the correct classes/subclasses

        if isinstance(key, str):
            if hasattr(self, key):
                return getattr(self, key)
            elif hasattr(self, '_observables') and key in self._observables.keys():
                # applicable only for Mesh, not ProtoMesh
                return self._observables[key]
            else:
                raise KeyError("{} is not a valid key".format(key))
        elif isinstance(key, np.ndarray):
            raise KeyError("use mesh.take(bool_per_triangle, bool_per_vertex) to split mesh")

    def __setitem__(self, key, value):
        """
        TODO: add documentation
        """

        # TODO: split this stuff into the correct classes/subclasses

        # NOTE: this will not be able to set values of ComputedColumns
        hkey = '_{}'.format(key)

        if hasattr(self, hkey):
            col = getattr(self, hkey)

            if isinstance(col, ComputedColumn) and not isinstance(value, ComputedColumn):
                col.set_for_computations(value)
            else:
                setattr(self, hkey, value)

            return

        elif hasattr(self, '_observables'):
            # applicable only for Mesh, not ProtoMesh
            # even if it doesn't exist, we'll make a new entry in observables]

            if key not in self._observables.keys():
                self._observables[key] = ComputedColumn(self)

            self._observables[key].set_for_computations(value)


        elif hasattr(self, key) and not hasattr(self, hkey):
            pass

        else:
            raise KeyError("{} is not a valid key".format(key))

    def keys(self):
        """
        TODO: add documentation
        """
        return self._keys

    def values(self):
        """
        TODO: add documentation
        """
        return self.items().values()

    def items(self):
        """
        TODO: add documentation
        """
        return {k: self[k] for k in self.keys()}

    def copy(self):
        """
        Make a deepcopy of this Mesh object
        """
        return copy.deepcopy(self)

    def take(self, bool_per_triangle, bool_per_vertex):
        """
        """
        if len(bool_per_triangle) != len(self.triangles):
            raise ValueError("bool_per_triangle should have length {}".format(len(self.triangles)))

        if len(bool_per_vertex) != len(self.vertices):
            raise ValueEror("bool_per_vertex should have length {}".format(len(self.vertices)))

        copy = self.copy()
        for k in self._keys:
            if not hasattr(copy, '_{}'.format(k)):
                # then likely a computed property which we don't need to set
                continue
            current_value = getattr(copy, k)

            if isinstance(current_value, float):
                # then something like volume/area which is no longer going
                # to be accurate by splitting the mesh, so we'll reset to None
                # copy[k] = None
                continue

            elif isinstance(current_value, bool):
                # then something like compute_at_vertices which we want to pass
                # on to each half
                copy[k] = current_value

            elif current_value is None or not len(current_value):
                # then this attribute is empty, so we can leave it that way
                # NOTE: we have to use == None instead of is None so that
                # computed columns can use their builtin __eq__ method
                continue
            elif len(current_value) == len(bool_per_triangle):
                copy[k] = current_value[bool_per_triangle]

                if k=='triangles':
                    # then we need to be clever since all the vertex indices will
                    # be re-ordered according to bool_per_vertex

                    # this is hideous, but I can't think of a clean way to do it,
                    # so for now I'll just make it work.

                    # We need to know the mapping for vertices from old to new index
                    index_mapping = {}  # index_mapping[old_index] = new_index
                    nv = 0
                    for i,b in enumerate(bool_per_vertex):
                        if b:
                            index_mapping[i] = nv
                            nv += 1

                    # now we need to apply this to all indices after filtering
                    try:
                        for it,triangle in enumerate(copy[k]):
                            for iv,vertex in enumerate(triangle):
                                copy[k][it][iv] = index_mapping[vertex]
                    except KeyError:
                        raise TypeError('Mesh failed or incomplete. Try increasing the number of triangles!')


            elif len(current_value) == len(bool_per_vertex):
                copy[k] = current_value[bool_per_vertex]
            else:
                raise ValueError("could not filter column {} (shape: {}) due to length mismatch (bool_per_triangle.shape: {}, bool_per_vertex.shape: {})".format(k, current_value.shape, bool_per_triangle.shape, bool_per_vertex.shape))

        return copy

    def update_columns_dict(self, kwargs):
        """
        Update the value of a column or multiple columns by passing as a dict.
        For observable columns, provide the label of the observable itself and
        it will be found (so long as it does not conflict with an existing
        non-observable column).
        """
        # make sure to do the geometric things that are needed for some of the
        # ComputedColumns first
        for key in ('triangles', 'vertices', 'centers', 'vnormals', 'tnormals'):
            if key in kwargs.keys():
                self.__setitem__(key, kwargs.pop(key))

        for k, v in kwargs.items():
            if isinstance(v, float) and k not in self._scalar_fields:
                # Then let's make an array with the correct length full of this
                # scalar

                # NOTE: this won't work for Nx3's, but that
                # really shouldn't ever happen since they should be set
                # within the init.
                # v = np.ones(self.Ntriangles)*v
                if self._compute_at_vertices:
                    v = np.full(self.Nvertices, v)
                else:
                    v = np.full(self.Ntriangles, v)

            self.__setitem__(k, v)

            if isinstance(v, ComputedColumn):
                # then let's update the mesh instance to correctly handle
                # inheritance
                self.__getitem__(k)._mesh = self


    def update_columns(self, **kwargs):
        """
        Update the value of a column or multiple columns by passing as kwargs.
        For observable columns, provide the label of the observable itself and
        it will be found (so long as it does not conflict with an existing
        non-observable column).
        """
        self.update_columns_dict(kwargs)

    @property
    def compute_at_vertices(self):
        """
        Access (read-only) to the setting of whether computations should
        be done at the vertices (and then averaged) or at the centers
        of each triangle

        :return: bool
        """
        return self._compute_at_vertices

    @property
    def Ntriangles(self):
        """
        Return the number of TRIANGLES/ELEMENTS in the mesh.

        Simply a shortcut to len(self.triangles)
        """
        return len(self.triangles)

    @property
    def Nvertices(self):
        """
        Return the number of VERTICES in the mesh.

        Simply a shortcut to len(self.vertices)
        """
        return len(self.vertices)


    @property
    def pvertices(self):
        """
        Return the array of vertices on the potential, where each item is a
        triplet representing cartesian coordinates.

        (Vx3)
        """
        return self._pvertices

    @property
    def vertices(self):
        """
        Return the array of vertices, where each item is a triplet
        representing cartesian coordinates.

        (Vx3)
        """
        return self._vertices

    @property
    def pvertices_per_triangle(self):
        """
        TODO: add documentation

        TODO: confirm shape
        (Nx3x3)
        """
        return self.pvertices[self.triangles]


    @property
    def vertices_per_triangle(self):
        """
        TODO: add documentation

        TODO: confirm shape
        (Nx3x3)
        """
        return self.vertices[self.triangles]

    @property
    def triangles(self):
        """
        Return the array of triangles, where each item is a triplet of indices
        pointing to vertices.

        (Nx3)
        """
        return self._triangles

    @property
    def centers(self):
        """
        Return the array of centers, where each item is a triplet representing
        cartesian coordinates.

        (Nx3)
        """
        return self._centers

    @property
    def coords_for_observations(self):
        """
        Return the coordinates from the center of the star for each element
        (either centers or vertices depending on the setting in the mesh)
        after perturbations (either by features or by offsetting to get
        the correct volume).  NOTE: this is NOT necessarily where the physical
        parameters were computed, but IS where eclipse detection, etc, is
        handled.
        """
        if self._compute_at_vertices:
            return self.vertices - self._pos_center
        else:
            return self.centers - self._pos_center

    @property
    def coords_for_computations(self):
        """
        Return the coordinates from the center of the star for each element
        (either centers or vertices depending on the setting in the mesh).
        """

        # TODO: need to subtract the position offset if a Mesh (in orbit)
        if self._compute_at_vertices:
            if self.pvertices is not None:
                return self.pvertices - self._pos_center
            else:
                return self.vertices - self._pos_center
        else:
            return self.centers - self._pos_center

    @property
    def normals_for_computations(self):
        """
        Return the normals for each element
        (either centers or vertices depending on the setting in the mesh).
        """
        if self._compute_at_vertices:
            return self.vnormals
        else:
            return self.tnormals

    @property
    def rs(self):
        """
        Return the radius of each element (either vertices or centers
        depending on the setting in the mesh) with respect to the center of
        the star.

        NOTE: unscaled

        (ComputedColumn)
        """
        rs = np.linalg.norm(self.coords_for_computations, axis=1)
        return ComputedColumn(self, rs)

    @property
    def rprojs(self):
        """
        Return the projected (in xy/uv plane) radius of each element (either
        vertices or centers depending on the setting in the mesh) with respect
        to the center of the star.

        NOTE: unscaled

        (ComputedColumn)
        """
        # TODO: should this be moved to Mesh?  Even though its surely possible
        # to compute without being placed in orbit, projecting in x,y doesn't
        # make much sense without LOS orientation.
        rprojs = np.linalg.norm(self.coords_for_computations[:,:2], axis=1)
        return ComputedColumn(self, rprojs)

    @property
    def cosbetas(self):
        """
        TODO: add documentation

        (ComputedColumn)
        """
        coords = self.coords_for_computations
        norms = self.normals_for_computations

        # TODO: ditch the list comprehension... I know I figured out how to do
        # this (ie along an axis) with np.dot somewhere else
        # cosbetas = np.array([np.dot(c,n) / (np.linalg.norm(c)*np.linalg.norm(n)) for c,n in zip(coords, norms)])

        cosbetas = libphoebe.scalproj_cosangle(
          np.ascontiguousarray(coords),
          np.ascontiguousarray(norms)
        )

        return ComputedColumn(self, cosbetas)

    @property
    def areas(self):
        """
        Return the array of areas, where each item is a scalar/float.

        TODO: UNIT?

        (Nx1)
        """
        return self._areas

    @property
    def tareas(self):
        """
        Return the array of triangle areas, where each item is a scalar/float.

        TODO: UNIT?

        (Nx1)
        """
        return self._tareas

    @property
    def areas_si(self):
        """
        TODO: add documentation
        """
        if self._areas is not None:
            return (self.areas*u.solRad**2).to(u.m**2).value
        else:
            return None


    @property
    def velocities(self):
        """
        Return the array of velocities, where each item is a scalar/float.

        TODO: UNIT?

        (Nx1)
        """
        return self._velocities

    @property
    def vnormals(self):
        """
        TODO: add documentation

        (Vx3)
        """
        return self._vnormals


    @property
    def tnormals(self):
        """
        Return the array of tnormals (normals for each triangle face), where
        each items is a triplet representing a cartesian normaled vector.

        (Nx3)
        """
        return self._tnormals

    @property
    def normgrads(self):
        """
        TODO: add documentation

        (ComputedColumn)
        """
        return self._normgrads

    @property
    def volume(self):
        """
        Return the volume of the ENTIRE MESH.

        (scalar/float)
        """
        return self._volume

    @property
    def area(self):
        """
        Return the surface area of the ENTIRE MESH.

        (scalar/float)
        """
        return self._area

    @property
    def phis(self):
        """
        TODO: add documentation
        """
        # TODO: if self._phis is None then compute from cartesian
        return self._phis

    @property
    def thetas(self):
        """
        TODO: add documentation
        """
        # TODO: if self._thetas is None then compute from cartesian
        return self._thetas

    @property
    def loggs(self):
        """
        Return the array of loggs, where each item is a scalar/float.

        (ComputedColumn)
        """
        return self._loggs

    @property
    def gravs(self):
        """
        Return the array of gravs, where each item is a scalar/float.

        TODO: UNIT?

        (ComputedColumn)
        """
        return self._gravs

    @property
    def teffs(self):
        """
        Return the array of teffs, where each item is a scalar/float.

        Unit: K

        (ComputedColum)
        """
        return self._teffs

    @property
    def abuns(self):
        """
        Return the array of abuns, where each item is a scalar/float.

        TODO: UNIT?

        (ComputedColumn)
        """
        return self._abuns

    @property
    def irrad_frac_refl(self):
        """
        Return the array of irrad_frac_refl, where each item is a scalar/float

        (ComputedColumn)
        """
        return self._irrad_frac_refl

    # @property
    # def frac_heats(self):
    #     """
    #     Return the array of frac_heats, where each item is a scalar/float

    #     (ComputedColumn)
    #     """
    #     return self._frac_heats

    # @property
    # def frac_scatts(self):
    #     """
    #     Return the array of frac_scatts, where each item is a scalar/float

    #     (ComputedColumn)
    #     """
    #     return self._frac_scatts




class ScaledProtoMesh(ProtoMesh):
    """
    ScaledProtoMesh is in real units (in whatever is provided for scale),
    but still with the origin at the COM of the STAR.

    Because there is no orbital or orientation information, those
    fields are not available until the mesh is placed in
    orbit (by using the class constructor on Mesh).
    """

    def __init__(self, **kwargs):
        """
        TODO: add documentation
        """

        # store needed copies in the Roche coordinates
        self._roche_vertices    = None  # Vx3
        self._roche_centers     = None  # Nx3
        self._roche_cvelocities = None  # Vx3
        self._roche_tnormals    = None  # Nx3

        keys = ['roche_vertices', 'roche_centers', 'roche_cvelocities', 'roche_tnormals']
        keys += kwargs.pop('keys', [])

        scale = kwargs.pop('scale', None)

        super(ScaledProtoMesh, self).__init__(keys=keys, **kwargs)

        if scale is not None:
            self._copy_roche_values()
            self._scale_mesh(scale)

    @classmethod
    def from_proto(cls, proto_mesh, scale):
        """
        TODO: add documentation
        """

        mesh = cls(**proto_mesh.items())
        mesh._copy_roche_values()
        mesh._scale_mesh(scale=scale)

        return mesh

    def _copy_roche_values(self):
        # make necessary copies
        # logger.debug("copying roche values")
        self._roche_vertices = self._vertices.copy()
        self._roche_centers = self._centers.copy()
        self._roche_cvelocities = self._velocities.centers.copy()
        self._roche_tnormals = self._tnormals.copy()

    def _scale_mesh(self, scale):
        """
        TODO: add documentation
        """
        pos_ks = ['vertices', 'centers']

        # TODO: scale velocities???

        # handle scale
        self.update_columns_dict({k: self[k]*scale for k in pos_ks})

        self.update_columns(areas=self.areas*(scale**2))
        self._volume *= scale**3
        if self._area is not None:
            # self._area is None for wd meshes
            self._area += scale**2

    @property
    def roche_coords_for_computations(self):
        """
        Return the coordinates from the center of the star for each element
        (either centers or vertices depending on the setting in the mesh).
        """

        if self._compute_at_vertices:
            return self.roche_vertices
        else:
            return self.roche_centers

    @property
    def roche_vertices(self):
        """
        Return the array of vertices in Roche coordinates, where each item is a
        triplet represeting cartesian coordinates.

        (Vx3)
        """
        return self._roche_vertices

    @property
    def roche_vertices_per_triangle(self):
        """
        TODO: add documentation

        TODO: confirm shape
        (Nx3x3)
        """
        return self.roche_vertices[self.triangles]

    @property
    def roche_centers(self):
        """
        Access to the quantities at the centers of each triangles in Roche
        coordinates.

        :return: numpy array
        """
        return self._roche_centers

    @property
    def roche_cvelocities(self):
        """
        Access to the velocities (compute at the centers of each triangle)
        in Roche coordinates

        :return: numpy array
        """
        return self._roche_cvelocities

    @property
    def roche_tnormals(self):
        """
        Access to the tnormals in Roche coordinates

        :return: numpy array
        """
        return self._roche_tnormals

class Mesh(ScaledProtoMesh):
    """
    Mesh is in real units (in whatever is provided for scale) with the
    origin at the COM of the system (the coordinate system that should be
    used when passing pos, vel, euler).

    Now that the mesh is in orbit and has viewing orientation information,
    observables can be attached.
    """
    def __init__(self, **kwargs):

        # self._mus               = None  # Nx1 property computed on-the-fly
        self._visibilities      = None  # Nx1
        self._weights           = None  # Nx3 (per vertex, same order as self.triangles)

        self._observables       = {}    # ComputedColumn (each)

        keys = ['mus', 'visibilities', 'weights', 'observables']
        keys = keys + kwargs.pop('keys', [])

        super(Mesh, self).__init__(keys=keys, **kwargs)

    @classmethod
    def from_proto(cls, proto_mesh, scale,
                      pos, vel, euler, euler_vel,
                      rotation_vel=(0,0,0),
                      component_com_x=None):
        """
        Turn a ProtoMesh into a Mesh scaled and placed in orbit.

        Update all geometry fields from the proto reference frame, to the
        current system reference frame, given the current position, velocitiy,
        euler angles, and rotational velocity of THIS mesh.

        :parameter list pos: current position (x, y, z)
        :parameter list vel: current velocity (vx, vy, vz)
        :parameter list euler: current euler angles (etheta, elongan, eincl)
        :parameter list rotation_vel: rotation velocity vector (polar_dir*freq_rot)
        """

        mesh = cls(**proto_mesh.items())

        mesh._copy_roche_values()
        mesh._scale_mesh(scale=scale)
        mesh._place_in_orbit(pos, vel, euler, euler_vel, rotation_vel, component_com_x)

        return mesh

    @classmethod
    def from_scaledproto(cls, scaledproto_mesh,
                         pos, vel, euler, euler_vel,
                         rotation_vel=(0,0,0),
                         component_com_x=None):
        """
        TODO: add documentation
        """

        mesh = cls(**scaledproto_mesh.items())

        # roche coordinates have already been copied
        # so do NOT call mesh._copy_roche_values() here
        mesh._place_in_orbit(pos, vel, euler, euler_vel, rotation_vel, component_com_x)

        return mesh

    def _place_in_orbit(self, pos, vel, euler, euler_vel, rotation_vel=(0,0,0), component_com_x=None):
        """
        TODO: add documentation
        """

        # TODO: store pos, vel, euler so that INCREMENTAL changes are allowed
        # if passing new values (and then make this a public method).  See note
        # below!

        pos_ks = ['vertices', 'pvertices', 'centers']
        norm_ks = ['vnormals', 'tnormals'] #, 'cnormals']
        vel_ks = ['velocities']

        # NOTE: we do velocities first since they require the positions WRT
        # the star (not WRT the system).  Will need to keep this in mind if we
        # eventually support incremental transformations.
        # pos_array = self.vertices if self._compute_at_vertices else self.centers
        pos_array = self.roche_vertices if self._compute_at_vertices else self.roche_centers
        if component_com_x is not None and component_com_x != 0.0:
            # then we're the secondary component and need to do 1-x and then flip the rotation component vxs
            pos_array = np.array([component_com_x, 0.0, 0.0]) - pos_array
        self.update_columns_dict({k: transform_velocity_array(self[k], pos_array, vel, euler_vel, rotation_vel) for k in vel_ks if self[k] is not None})
        # TODO: handle velocity from mesh reprojection during volume conservation

        # handle rotation/displacement
        # NOTE: mus will automatically be updated on-the-fly
        self.update_columns_dict({k: transform_position_array(self[k], pos, euler, False) for k in pos_ks if self[k] is not None})
        self.update_columns_dict({k: transform_position_array(self[k], pos, euler, True) for k in norm_ks if self[k] is not None})

        # let's store the position.  This is both useful for "undoing" the
        # orbit-offset, and also eventually to allow incremental changes.
        self._pos = pos
        if component_com_x is not None and component_com_x != 0.0:
            self._pos_center = transform_position_array(np.array([component_com_x, 0.0, 0.0]), pos, euler, False)
        else:
            self._pos_center = pos
        self._euler = euler


    @property
    def mus(self):
        """
        Return the array of mus (z-coordinate of tnormals), where each item
        is a scalar/float.

        (Nx1)
        """
        # this requires tnormals to be NORMALIZED (currently we don't do any
        # checks - this is just assumed)
        return self.tnormals[:,2]

    @property
    def mus_for_computations(self):
        """
        TODO: add documentation

        (Nx1 or Vx1)
        """
        if self._compute_at_vertices:
            # this requires vnormals to be NORMALIZED (assumed)
            return self.vnormals[:,2]
        else:
            # this is just self.tnormals[:,2] from above
            return self.mus

    @property
    def visibilities(self):
        """
        Return the array of visibilities, where each item is a scalar/float
        between 0 (completely hidden/invisible) and 1 (completely visible).

        (Nx1)
        """
        if self._visibilities is not None:
            return self._visibilities
        else:
            return np.ones(self.Ntriangles)

    @property
    def weights(self):
        """
        TODO: add documentation

        (Nx3)
        """
        if self._weights is not None and len(self._weights):
            return self._weights
        else:
            return np.full((self.Ntriangles, 3), 1./3)

    @property
    def observables(self):
        """
        Return the dictionary of observables

        (ComputedColumn)
        """
        return self._observables

    def get_observable(self, label):
        """
        Retrieve the array of an observable by its label.

        (ComputedColumn)
        """
        return self.observables[label]

    def fill_observable(self, label, value):
        """
        Fill the array of an observable by its label.  The provided value
        MUST be either a scalar (applied for all triangles), or an array
        with length N.  This IS NOT checked, but failing to provide the
        correct shape could have dire consequences.
        """
        self._observables[label] = value

    def update_columns_dict(self, kwargs):
        """
        TODO: add documentation
        """
        super(Mesh, self).update_columns_dict(kwargs)

        # if kwargs.get('vnormals', None) is not None or kwargs.get('tnormals', None) is not None:
            # self._compute_mus()
        if kwargs.get('triangles', None) is not None:
            # reset visibilities and velocities so that they are reset
            # when next queried
            self.update_columns(visibilities=None, velocities=None)


class Meshes(object):
    """
    This is just a class to allow easy access to meshes at the system-level.
    This allows eclipse detection and subdivision to work at the system-level in
    an efficient and clean way.

    In effect this has the same structure as the system, but allows passing only
    the meshes and dropping all the Body methods/attributes.
    """
    def __init__(self, items, parent_envelope_of={}):
        """
        TODO: add documentation

        :parameter list items:
        """
        items = {k:v for k,v in items.items() if v is not None}
        if isinstance(list(items.values())[0], ProtoMesh):
            self._dict = items
        else:
            self._dict = {component: body.mesh for component, body in items.items()}
        #self._component_by_no = {body.comp_no: component for component, body in items.items()}
        self._components = list(items.keys())
        self._parent_envelope_of = parent_envelope_of

    def items(self):
        """
        TODO: add documentation
        """
        return self._dict.items()

    def keys(self):
        """
        TODO: add documentation
        """
        return list(self._dict.keys())

    def values(self):
        """
        TODO: add documentation
        """
        return list(self._dict.values())

    def __getitem__(self, key):
        """
        TODO: add documentation
        """
        if key in self._components:
            return self._dict[key]
        else:
            return self.get_column_flat(key, self._components)

    def __setitem__(self, key, value):
        """
        TODO: add documentation
        """
        return self.update_columns(key, value)

    @property
    def Ntriangles(self):
        return sum(mesh.Ntriangles for mesh in self.values())

    @property
    def Nvertices(self):
        return sum(mesh.Nvertices for mesh in self.values())

    def __getattr__(self, attr):
        if np.all([hasattr(mesh, attr) for mesh in self.values()]):
            # need to be smart if this is supposed to return a ComputedColumn...
            items = [getattr(mesh, attr) for mesh in self.values()]
            if isinstance(items[0], ComputedColumn):
                return ComputedColumn(self, np.concatenate([item.for_computations for item in items]))
            elif isinstance(items[0], bool):
                # this may be a bit of an assumption...
                return np.all(items)
            elif isinstance(items[0], np.ndarray):
                return np.concatenate(items)
            else:
                # for floats we probably always want the sum... but instead
                # we'll define those explicitly above.  Anything that isn't
                # covered above should then raise an exception here.
                raise NotImplementedError()
        else:
            raise AttributeError("meshes don't have attribute {}".format(attr))

    def component_by_no(self, comp_no):
        """
        TODO: add documentation
        """
        #return self._component_by_no[comp_no]
        return self._components[comp_no-1]

    def update_columns(self, field, value_dict, inds=None, computed_type=None):
        """
        update the columns of all meshes

        :parameter str field: name of the mesh columnname
        :parameter value_dict: dictionary with component as keys and new
            data as values. If value_dict is not a dictionary,
            it will be applied to all components
        :type value_dict: dict or value (array or float)
        """
        if not isinstance(value_dict, dict):
            value_dict = {comp_no: value_dict for comp_no in self._dict.keys()}


        for comp, value in value_dict.items():
            if computed_type is not None:
                # then create the ComputedColumn now to override the default value of compute_at_vertices
                self._dict[comp]._observables[field] = ComputedColumn(self._dict[comp], compute_at_vertices=computed_type=='vertices')

            #print "***", comp, field, inds, value
            if inds:
                raise NotImplementedError('setting column with indices not yet ported to new meshing')
                # self._dict[comp][field][inds] = value
            else:
                if comp in self._dict.keys():
                    if self._dict[comp] is None:
                        return
                    self._dict[comp][field] = value
                else:
                    meshes = self._dict[self._parent_envelope_of[comp]]
                    meshes[comp][field] = value

    def get_column(self, field, components=None, computed_type='for_observations'):
        """
        TODO: add documentation

        return a dictionary for a single column, with component as keys and the
        column array as values

        :parameter str field: name of the mesh columnname
        :parameter components:
        """
        def get_field(c, field, computed_type):

            if c not in self._dict.keys() and self._parent_envelope_of[c] in self._dict.keys():
                mesh = self._dict[self._parent_envelope_of[c]]
                return mesh.get_column_flat(field, components, computed_type)

            mesh = self._dict[c]
            if isinstance(mesh, Meshes):
                # then do this recursively for all components in the Meshes object
                # but don't allow nesting in the dictionary, instead combine
                # all subcomponents into one entry with the current component
                return mesh.get_column_flat(field, mesh._components, computed_type)
            elif mesh is None:
                return None

            f = mesh[field]
            if isinstance(f, ComputedColumn):
                col = getattr(f, computed_type)
            else:
                col =  f

            return col

        if components:
            if isinstance(components, str):
                components = [components]
        else:
            components = list(self.keys())

        return {c: get_field(c, field, computed_type) for c in components}

    def get_column_flat(self, field, components=None, computed_type='for_observations'):
        """
        TODO: add documentation

        return a single merged value (hstacked) from all meshes

        :parameter str field: name of the mesh columnname
        :parameter components:
        """
        return self.pack_column_flat(self.get_column(field, components, computed_type),
                                     components,
                                     offset=field=='triangles')

    def pack_column_flat(self, value, components=None, offset=False):
        """
        TODO: add documentation
        """

        if components:
            if isinstance(components, str):
                components = [components]
            elif isinstance(components, list):
                components = components
            else:
                raise TypeError("components should be list or string, not {}".format(type(components)))
            value = {c: v for c,v in value.items() if v is not None}
            components = list(value.keys())
        elif isinstance(value, dict):
            value = {c: v for c,v in value.items() if v is not None}
            components = list(value.keys())
        elif isinstance(value, list):
            value = {c: v for c,v in zip(list(self._dict.keys()), value) if v is not None}
            components = list(value.keys())

        if not len(components):
            return np.array([])

        if offset:
            values = []
            offsetN = 0
            for c in components:
                values.append(value[c]+offsetN)
                offsetN += len(self[c]['vertices'])

        else:
            values = [value[c] for c in components]

        if len(value[components[0]].shape) > 1:
            return np.vstack(values)
        else:
            return np.hstack(values)

    def unpack_column_flat(self, value, components=None, offset=False, computed_type=None):
        """
        TODO: add documentation
        TODO: needs testing
        """
        if components:
            if isinstance(components, str):
                components = [components]
        else:
            components = self._dict.keys()

            # TODO: add this

        # we need to split the flat array by the lengths of each mesh
        N_lower = 0
        N_upper = 0
        offsetN = 0.0
        value_dict = {}
        for comp in components:
            if isinstance(self[comp], Meshes):
                # then we need to recursively extract to the underlying meshes
                # pass
                meshes = self[comp]._dict
            else:
                meshes = {comp: self[comp]}

            for c, mesh in meshes.items():
                if mesh is None:
                    continue
                if computed_type=='vertices' or (computed_type is None and mesh._compute_at_vertices):
                    N = mesh.Nvertices
                else:
                    N = mesh.Ntriangles
                N_upper += N
                value_dict[c] = value[N_lower:N_upper] - offsetN
                if offset:
                    offsetN += N
                N_lower += N

        return value_dict

    def set_column_flat(self, field, value, components=None, computed_type=None):
        """
        TODO: add documentation
        TODO: needs testing
        """
        value_dict = self.unpack_column_flat(value, components, computed_type=computed_type)
        self.update_columns(field, value_dict, computed_type=computed_type)

    def replace_elements(self, inds, new_submesh, component):
        """
        TODO: add documentation
        TODO: remove this method???
        """
        self._dict[component] = np.hstack([self._dict[component][~inds], new_submesh])



import numpy as np
import astropy.units as u

import sys
import os

if sys.version_info[0] < 3:
    from urllib2 import urlopen as _urlopen
else:
    from urllib.request import urlopen as _urlopen

import json as _json

try:
  basestring = basestring
except NameError:
  basestring = str

class Alias(object):
    def __init__(self, dict):
        self._dict = dict

    def map(self, key):
        return self._dict.get(key, key)


# https://matplotlib.org/examples/color/named_colors.html
# technically some of these (y, c, and m) don't exactly match, but as they "mean"
# the same thing, we'll handle the alias anyways
# Convention: always map to the spelled name and gray over grey
coloralias = Alias({'k': 'black', 'dimgrey': 'dimgray', 'grey': 'gray',
              'darkgrey': 'darkgray', 'lightgrey': 'lightgray',
               'w': 'white', 'r': 'red', 'y': 'yellow', 'g': 'green',
               'c': 'cyan', 'b': 'blue', 'm': 'magenta'})

# Convention: always map to the spelled name
linestylealias = Alias({'_': 'solid', '--': 'dashed', ':': 'dotted',
                      '-.': 'dashdot'})

class Group(object):
    def __init__(self, baseclass, reprlist, items):
        if not isinstance(items, list):
            raise TypeError("items must be of type list of {} objects".format(baseclass.__name__))

        # handle the case of unpacking nested groups (NOTE: only 1 deep)
        items_flatten = []
        for item in items:
            if isinstance(item, Group):
                for groupitem in item._items:
                    items_flatten.append(groupitem)
            else:
                items_flatten.append(item)

        for item in items_flatten:
            if not isinstance(item, baseclass):
                raise TypeError("each item in items must be of type {}".format(baseclass.__name__))

        self._baseclass = baseclass
        self._items = items_flatten
        self._reprlist = reprlist

    def __repr__(self):
        info = " | ".join(["{}s: {}".format(attr,
                                           ", ".join(getattr(self, attr)))
                           for attr in self._reprlist])

        return "<{} | {} items | {}>".format(self.__class__.__name__, len(self), info)

    # @classmethod
    # def from_dict(cls, dict):
    #     return cls(**dict)
    #
    # def to_dict(self):
    #     return {'baseclass': self._baseclass,
    #             'items': self._items,
    #             'reprlist': self._reprlist}

    def __getitem__(self, ind):
        return self._items.__getitem__(ind)

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    def _get_attrs(self, attr):
        return [getattr(d, attr) for d in self._items]

    def _set_attrs(self, attr, value):
        for d in self._items:
            setattr(d, attr, value)


def _convert_unit(unit):
    if unit is None:
        unit = u.dimensionless_unscaled

    if isinstance(unit, basestring):
        unit = u.Unit(unit)

    if not (isinstance(unit, u.Unit) or isinstance(unit, u.CompositeUnit) or isinstance(unit, u.IrreducibleUnit)):
        raise TypeError("unit must be of type Unit, got: {}".format(unit))

    return unit

def tolist(value):
    if isinstance(value, np.ndarray):
        return value.tolist()
    else:
        return [value]

def arraytolistrecursive(value):
    if hasattr(value, '__iter__') and not isinstance(value, str):
        return [arraytolistrecursive(v) for v in value]
    else:
        return value

def _bytes(s):
    if sys.version_info[0] == 3:
        return bytes(s, 'utf-8')
    else:
        return bytes(s)

def _json_safe(item):
    if isinstance(item, np.ndarray):
        return arraytolistrecursive(item)
    elif isinstance(item, dict):
        return {k: _json_safe(v) for k,v in item.items()}
    else:
        return item

def _parse_json(pairs):
    """
    modified from:
    https://stackoverflow.com/questions/956867/how-to-get-string-objects-instead-of-unicode-from-json#34796078

    pass this to the object_pairs_hook kwarg of json.load/loads
    """
    def _string(item):
        if isinstance(item, bytes):
            # return item.decode('utf-8')
            return _bytes(item)
        elif sys.version_info[0] == 2 and isinstance(item, unicode):
            return item.encode('utf-8')
        else:
            return item

    new_pairs = []
    for key, value in pairs:
        key = _string(key)

        if isinstance(value, dict):
            value = _parse_json(value.items())
        elif isinstance(value, list):
            value = [_string(v) for v in value]
        else:
            value = _string(value)

        new_pairs.append((key, value))
    return dict(new_pairs)

def save(dict, filename):
    filename = os.path.expanduser(filename)
    f = open(filename, 'w')
    _json.dump(dict, f,
              sort_keys=False, indent=0)

    f.close()

    return filename

def load(filename):
    if filename[:4] == 'http':
        resp = _urlopen(filename)
        dict = _json.loads(resp.read(), object_pairs_hook=_parse_json)
    else:
        filename = os.path.expanduser(filename)
        f = open(filename, 'r')
        dict = _json.load(f, object_pairs_hook=_parse_json)
        f.close()

    return dict

dimensions = ['i', 'x', 'y', 'z', 's', 'c']

global _inline
_inline = False



import sys
import numpy as np

from phoebe.algorithms import ceclipse
from phoebe.utils import _bytes
import libphoebe

import logging

_EXPORT_HORIZON = False

logger = logging.getLogger("ECLIPSE")

def _graham_scan_inside_hull(front, back):
    sa = np.argsort(front[:,0], kind='heapsort')
    hull, inside = ceclipse.graham_scan_inside_hull(front[sa], back)
    return hull, inside

"""
each of these functions needs to take meshes, xs, ys, zs.
- meshes is a dictionary with keys being the component number of that mesh and
  values being the meshes themselves
- xs, ys, zs are lists of the positions of each component (ie xs[comp_no-1])
each function must return a dictionary, with keys being component numbers and values
  being the revised visibilities of the corresponding mesh

"""

def wd_horizon(meshes, xs, ys, zs, expose_horizon=False):
    """
    """

    nbodies = len(meshes.keys())

    if nbodies != 2:
        raise ValueError("wd_horizon eclipse method only works on binaries")

    front_to_back_comp_nos = np.argsort(zs)[::-1]
    i_front = front_to_back_comp_nos[0]
    i_back = front_to_back_comp_nos[1]

    comp_front = meshes.component_by_no(i_front+1)
    comp_back = meshes.component_by_no(i_back+1)

    mesh_front = meshes[comp_front]
    mesh_back = meshes[comp_back]


    # polar coordinates need to be wrt the center of the ECLIPSING body
    rhos = [np.sqrt((mesh.centers[:,0]-xs[i_front])**2 +
                    (mesh.centers[:,1]-ys[i_front])**2) for mesh in (mesh_front, mesh_back)]
    # thetas = [np.arcsin((mesh['center'][:,1]-ys[i_front])/rho) for mesh,rho in zip((mesh_front, mesh_back), rhos)]
    thetas = [np.arctan2(mesh.centers[:,1]-ys[i_front], mesh.centers[:,0]-xs[i_front]) for mesh in (mesh_front, mesh_back)]
    # mus = [mesh['mu'] for mesh in (mesh_front, mesh_back)]

    # import matplotlib.pyplot as plt
    # plt.plot(rhos[0], thetas[0], 'b.')
    # plt.plot(rhos[1], thetas[1], 'r.')
    # plt.show()

    # To find the horizon we want the first positive mu elements (on right and
    # left) for each latitude strip.

    horizon_inds = []
    # we only need the horizon of the ECLIPSING star, so we'll use mesh_front
    # and mus[i_front], xs[i_front], etc
    lats = list(set(mesh_front.theta))
    for lat in lats:
        lat_strip_inds = mesh_front.theta == lat

        # let's get the x-coordinate wrt THIS star so we can do left vs right
        x_rel = mesh_front.centers[:,0] - xs[i_front]

        print(lat, x_rel[lat_strip_inds].min(), x_rel[lat_strip_inds].max())

        # and since we want the first element in the front, let's just get rid of the back
        front_inds = mesh_front.mus >= 0.0
        back_inds = mesh_front.mus < 0.0

        left_inds = x_rel < 0.0
        right_inds = x_rel >= 0.0

        # let's handle "left" vs "right" side of star separately
        for side_inds, fb_inds, side in zip((left_inds, right_inds), (front_inds, back_inds), ('left', 'right')):
            no_triangles = len(mesh_front.mus[lat_strip_inds * side_inds * fb_inds])
            # print "*** no triangles at lat", lat, no_triangles

            if no_triangles > 0:
                if side=='left':
                    # then we want the first triangle on the FRONT of the star
                    first_horizon_mu = mesh_front.mus[lat_strip_inds * side_inds * fb_inds].min()
                else:
                    # then we want the first triangle on the BACK of the star
                    first_horizon_mu = mesh_front.mus[lat_strip_inds * side_inds * fb_inds].max()

                first_horizon_ind = np.where(mesh_front.mus==first_horizon_mu)[0][0]
                # print "*** horizon index", first_horizon_ind

                horizon_inds.append(first_horizon_ind)

    thetas[0][thetas[0] < 0] = thetas[0][thetas[0] < 0]+2*np.pi

    f = open('bla2', 'w')
    for ind in horizon_inds:
        # note here rhos and thetas need 0 not i_front because of the stupid way we did the list comprehension
        f.write("{} {} {}\n".format(ind, rhos[0][ind], thetas[0][ind]))
    f.close()

    # these are the horizon coordinates for the ECLIPSING star
    rhos[0][horizon_inds]
    thetas[0][horizon_inds]

    visibilities, weights, horizon = only_horizon(meshes, xs, ys, zs)
    # now edit visibilities based on eclipsing region
    # visibilities = {comp_no: np.ones(len(mesh)) for comp_no, mesh in meshes.items()}
    visibilities[comp_back]
    mesh_back

    return visibilities, None, None


def none(meshes, xs, ys, zs, expose_horizon=False):
    """
    """
    return {comp_no: mesh.visibilities for comp_no, mesh in meshes.items()}, None, None

def only_horizon(meshes, xs, ys, zs, expose_horizon=False):
    """
    Check all visible or partial triangles to see if they're behind the horizon,
    by checking the direction of the z-component of the normals (ie hidden if mu<0)
    """

    # if visibility == 0, it should remain 0
    # if visibility == 0.5, it should stay 0.5 if mu > 0 else it should become 0
    # if visibility == 1, it should stay 1 if mu > 0 else it should become 0

    # this can all by easily done by multiplying by int(mu>0) (1 if visible, 0 if hidden)

    return {comp_no: mesh.visibilities * (mesh.mus > 0).astype(int) for comp_no, mesh in meshes.items() if mesh is not None}, None, None

def native(meshes, xs, ys, zs, expose_horizon=False, horizon_method='boolean'):
    """
    TODO: add documentation

    this is the new eclipse detection method in libphoebe
    """

    centers_flat = meshes.get_column_flat('centers')
    vertices_flat = meshes.get_column_flat('vertices')
    triangles_flat = meshes.get_column_flat('triangles')  # should handle offset automatically

    if horizon_method=='boolean':
        normals_flat = meshes.get_column_flat('tnormals')
    elif horizon_method=='linear':
        normals_flat = meshes.get_column_flat('vnormals')
    else:
        raise NotImplementedError

    # viewing_vector is defined as star -> earth
    # NOTE: this will need to flip if we change the convention on the z-direction
    viewing_vector = np.array([0., 0., 1.])


    # we need to send in ALL vertices but only the visible triangle information
    info = libphoebe.mesh_visibility(viewing_vector,
                                     vertices_flat,
                                     triangles_flat,
                                     normals_flat,
                                     tvisibilities=True,
                                     taweights=True,
                                     method=_bytes(horizon_method),
                                     horizon=expose_horizon)

    visibilities = meshes.unpack_column_flat(info['tvisibilities'], computed_type='triangles')
    weights = meshes.unpack_column_flat(info['taweights'], computed_type='triangles')

    if expose_horizon:
        horizons = info['horizon']

        # TODO: we need to do this per component and somehow return them in
        # a predictable order or as a dictionary a la the other returned quantities
        horizons = [vertices_flat[horizon_i] for horizon_i in horizons]

    else:
        horizons = None

    return visibilities, weights, horizons

def visible_partial(meshes, xs, ys, zs, expose_horizon=False):
    centers_flat = meshes.get_column_flat('centers')
    vertices_flat = meshes.get_column_flat('vertices')
    triangles_flat = meshes.get_column_flat('triangles')  # should handle offset automatically
    normals_flat = meshes.get_column_flat('tnormals')

    # viewing_vector is defined as star -> earth
    # NOTE: this will need to flip if we change the convention on the z-direction
    viewing_vector = np.array([0., 0., 1.])

    visibilities = libphoebe.mesh_rough_visibility(viewing_vector,
                                                   vertices_flat,
                                                   triangles_flat,
                                                   normals_flat)

    visibilities = meshes.unpack_column_flat(visibilities, computed_type='triangles')

    return visibilities, None, None



def graham(meshes, xs, ys, zs, expose_horizon=False):
    """
    convex_graham
    """

    distance_factor = 1.0  # TODO: make this an option (what does it even do)?

    # first lets handle the horizon
    visibilities, weights, horizon = only_horizon(meshes, xs, ys, zs)

    # Order the bodies from front to back.  We can do this through zs (which are
    # the z-coordinate positions of each body in the system.
    # Those indices are in the same order as the meshes['comp_no']
    # (needing to add 1 because comp_no starts at 1 not 0)

    # TODO: I don't think this whole comp_no thing is going to work with nested
    # triples.. may need to rethink the whole meshes.component_by_no thing

    front_to_back_comp_nos = np.argsort(zs)[::-1]+1
    nbodies = len(front_to_back_comp_nos)

    for i_front in range(0, nbodies-1):
        # for a binary, i_front will only be 0
        comp_no_front = front_to_back_comp_nos[i_front]
        comp_front = meshes.component_by_no(comp_no_front)
        mesh_front = meshes[comp_front]
        visibility_front = visibilities[comp_front]

        for i_back in range(i_front+1, nbodies):
            # for a binary, i_back will only be 1
            comp_no_back = front_to_back_comp_nos[i_back]
            comp_back = meshes.component_by_no(comp_no_back)
            mesh_back = meshes[comp_back]
            visibility_back = visibilities[comp_back]

            # If mesh_back is entirely hidden, then we can skip any checks and
            # leave it hidden.  Note that here we need to check visibility instead
            # of mesh['visibility'] or mesh_back['visibility'] since we are not
            # adjusting those values in memory.
            if np.all(visibility_back==0.0):
                continue

            # Determine a scale factor for the triangle
            min_size_back = mesh_back.areas.min()
            distance = distance_factor * 2.0/3**0.25*np.sqrt(min_size_back)

            # Select only those triangles that are not hidden
            tri_back_vis = mesh_back.vertices_per_triangle[visibility_back > 0.0].reshape(-1,9)
            tri_front_vis = mesh_front.vertices_per_triangle[visibility_front > 0.0].reshape(-1,9)

            back = np.vstack([tri_back_vis[:,0:2], tri_back_vis[:,3:5], tri_back_vis[:,6:8]])
            front = np.vstack([tri_front_vis[:,0:2], tri_front_vis[:,3:5], tri_front_vis[:,6:8]])

            # Star in front ---> star in back
            if not front.shape[0]:
                continue

            hull, inside = _graham_scan_inside_hull(front, back)
            hidden = inside.reshape(3,-1).all(axis=0)
            visible = ~(inside.reshape(3,-1).any(axis=0))

            # Triangles that are partially hidden are those that are not
            # completely hidden, but do have at least one vertex hidden
            partial = ~hidden & ~visible

            # These returned visibilities are only from mesh_back_vis
            # So to apply back to our master visibilities parameter, we need
            # to find the correct inds.
            visibility_back[visibility_back > 0.0] = 1.0*visible + 0.5*partial


            ###################################################
            ## TODO: port the following code to be included  ##
            ###################################################


            # It's possible that some triangles that do not have overlapping
            # vertices, still are partially covered (i.e. if the objects fall
            # in between the vertices)
            #visible1 = ~star1.mesh['hidden']
            #backc = star1.mesh['center'][visible1,0:2]
            #if not len(backc):
            #    continue
            #
            #subset_partial = closest_points(hull, backc, distance)
            #
            #if not len(subset_partial):
            #    continue
            #
            #arg = np.where(visible1)[0][subset_partial]
            #star1.mesh['visible'][arg] = False
            #star1.mesh['partial'][arg] = True


    return visibilities, None, None



import numpy as np
from astropy import units as u
from scipy.optimize import newton, minimize, least_squares
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter
from scipy.version import version as _scipy_version
from phoebe.constraints.builtin import t0_supconj_to_perpass
from copy import deepcopy
from distutils.version import LooseVersion

# if os.getenv('PHOEBE_ENABLE_PLOTTING', 'TRUE').upper() == 'TRUE':
#     try:
#         import matplotlib.pyplot as plt
#     except (ImportError, TypeError):
#         _use_mpl = False
#     else:
#         _use_mpl = True
# else:
#     _use_mpl = False


def smooth_rv(rvdata):
    if rvdata is None:
        return None

    win_len = int(len(rvdata)/5)
    win_len = int(win_len + 1) if int(win_len)%2 == 0 else int(win_len)
    win_len = 5 if win_len <= 3 else win_len
    poly_ord = 3
    rv_smooth = savgol_filter(rvdata[:,1], window_length=win_len, polyorder=poly_ord)
 
    return np.array([rvdata[:,0], rv_smooth, rvdata[:,2]]).T

def interpolate_rvs(rv1data, rv2data):

    phases1, rvs1 = rv1data[:,0], rv1data[:,1]
    phases2, rvs2 = rv2data[:,0], rv2data[:,1]
    # find the min and max phases of both arrays
    ph1_min, ph1_max = phases1.min(), phases1.max()
    ph2_min, ph2_max = phases2.min(), phases2.max()

    # determine the overlapping region
    ph_min = max([ph1_min, ph2_min])
    ph_max = min([ph1_max, ph2_max])

    # filter the rv arrays to only include the overlapping regions
    # interpolate in the array that covers more phases

    cond1 = (ph_min <= phases1) & (phases1 <= ph_max)
    cond2 = (ph_min <= phases2) & (phases2 <= ph_max)

    ph1_new, rvs1_new = phases1[cond1], rvs1[cond1]
    ph2_new, rvs2_new = phases2[cond2], rvs2[cond2]

    if len(ph1_new) > len(ph2_new):
        # interpolate rv2 in ph1 phases
        rv2_interp = interp1d(ph2_new, rvs2_new)
        rvs2_new = rv2_interp(ph1_new)
    else:
        # interpolate rv1 in ph2 phases
        rv1_interp = interp1d(ph1_new, rvs1_new)
        rvs1_new = rv1_interp(ph2_new)

    return rvs1_new, rvs2_new
    

def estimate_q(rv1data, rv2data, vgamma=None, tol=1e-8):
    if len(rv1data) != len(rv2data):
        rvs1, rvs2 = interpolate_rvs(rv1data, rv2data)
    elif (np.abs(rv1data[:,0] - rv2data[:,0]) > 1e-8).any():
        rvs1, rvs2 = interpolate_rvs(rv1data, rv2data)
    else:
        rvs1 = rv1data[:,1]
        rvs2 = rv2data[:,1]

    rv1max = max(rvs1)
    rv2min = rvs2[rvs1 ==rv1max]
#     rv2max = max(rv2data[:,1])
    if vgamma is None:
        return np.abs(rv1max/rv2min)
    else:
        return (rv1max-vgamma)/(-rv2min+vgamma)


def estimate_vgamma(rv1data, rv2data, q=1.):
    if len(rv1data) != len(rv2data):
        rvs1, rvs2 = interpolate_rvs(rv1data, rv2data)
    elif (np.abs(rv1data[:,0] - rv2data[:,0]) > 1e-8).any():
        rvs1, rvs2 = interpolate_rvs(rv1data, rv2data)
    else:
        rvs1 = rv1data[:,1]
        rvs2 = rv2data[:,1]

    return np.mean(rvs1+q*rvs2)/(1+q)


def estimate_q_vgamma(rv1data, rv2data, maxiter=10):
    if rv2data is None:
        rv1_flipped = np.array([rv1data[:,0],
                                -rv1data[:,1] + rv1data[:,1].max() + rv1data[:,1].min(),
                                rv1data[:,2]]).T
        return np.nan, estimate_vgamma(rv1data, rv1_flipped, q=1.)

    if rv1data is None:
        rv2_flipped = np.array([rv2data[:,0],
                                -rv2data[:,1] + rv2data[:,1].max() + rv2data[:,1].min(),
                                rv2data[:,2]]).T
        return np.nan, estimate_vgamma(rv2data, rv2_flipped, q=1.)


    q_prev = estimate_q(rv1data, rv2data, vgamma=None)
    vgamma_prev = estimate_vgamma(rv1data, rv2data, q_prev)

    for i in range(maxiter):
        q_est = estimate_q(rv1data, rv2data, vgamma=vgamma_prev)
        vgamma_est = estimate_vgamma(rv1data, rv2data, q_est)

        if np.abs((q_est - q_prev)/q_est) < 1e-2 and np.abs((vgamma_est -  vgamma_prev)/vgamma_est) < 1e-2:
            break

        q_prev = q_est
        vgamma_prev = vgamma_est

    return q_est[0], vgamma_est[0]


def estimate_asini(rv1data, rv2data, period = 1*u.d, vgamma = 0., ecc=0.):
    period  = (period.to(u.s)).value

    K1 = 0.5*(max(rv1data[:,1]-vgamma)-min(rv1data[:,1]-vgamma)) if rv1data is not None else np.nan
    asini1 = K1*period*(1-ecc**2)**0.5/(2*np.pi)
    K2 = 0.5*(max(rv2data[:,1]-vgamma)-min(rv2data[:,1]-vgamma)) if rv2data is not None else np.nan
    asini2 = K2*period*(1-ecc**2)**0.5/(2*np.pi)

    return [asini1, asini2]


def estimate_phase_supconj(rv1data, rv2data, vgamma):

    if rv1data is not None:
        grad1 = np.gradient(rv1data[:,1]) if rv1data is not None else np.nan
        ph_rv1_interp = interp1d(rv1data[:,1][grad1<=0], rv1data[:,0][grad1<=0])
        ph_vgamma_1 = ph_rv1_interp(vgamma)
    else:
        ph_vgamma_1 = np.nan

    if rv2data is not None:
        grad2 = np.gradient(rv2data[:,1]) if rv2data is not None else np.nan
        ph_rv2_interp = interp1d(rv2data[:,1][grad2>=0], rv2data[:,0][grad2>=0])
        ph_vgamma_2 = ph_rv2_interp(vgamma)
    else:
        ph_vgamma_2 = np.nan

    return np.nanmean([ph_vgamma_1, ph_vgamma_2])


def ecc_anomaly(x, phases, ph0, ecc):
    return x-ecc*np.sin(x) - 2*np.pi*(phases-ph0)


def rv_model(phases, P, per0, ecc, asini, vgamma, ph_supconj, component=1):

    ph0 = t0_supconj_to_perpass(ph_supconj, 1., ecc, per0, 0., 0., 0.)
    # deepcopy is needed for scipy < 1.2.2 because of this bug: https://github.com/scipy/scipy/issues/9964
    Es = newton(ecc_anomaly,
                deepcopy(phases) if LooseVersion(_scipy_version) < LooseVersion("1.2.2") else phases,
                args=(phases, ph0*np.ones_like(phases), ecc*np.ones_like(phases)))

    thetas = 2*np.arctan(((1+ecc)/(1-ecc))**0.5*np.tan(Es/2))
    P_s = ((P*u.d).to(u.s)).value
    if component==1:
        const = 2*np.pi*asini[0]/(P_s*(1-ecc**2)**0.5)
    elif component==2:
        const = -2*np.pi*asini[1]/(P_s*(1-ecc**2)**0.5)
    else:
        raise ValueError('Unrecognized component %i, can only be 1 or 2' % (component))

    tdep = ecc*np.cos(per0)+np.cos(per0+thetas)
    return (const*tdep) + vgamma


def loglike(params, rv1data, rv2data, asini, vgamma, ph_supconj):
    logl1 = 0
    logl2 = 0

    ecc, per0 = params
    period = 1. # because phase-folded rv
    if rv1data is not None:
        rvs1 = rv_model(rv1data[:,0], period, per0, ecc, asini, vgamma, ph_supconj,component=1)
        logl1 = -0.5*np.sum((rv1data[:,1]-rvs1)**2/(rv1data[:,2])**2)

    if rv2data is not None:
        rvs2 = rv_model(rv2data[:,0], period, per0, ecc, asini, vgamma, ph_supconj, component=2)
        logl2 = -0.5*np.sum((rv2data[:,1]-rvs2)**2/(rv2data[:,2])**2)
#     print(logl1+logl2)
    return logl1+logl2


def estimate_rv_parameters(rv1data=None, rv2data=None,
                           q=None, vgamma=None, asini=None, ecc=None, per0=None, maxiter=10):

    rv1_smooth = smooth_rv(rv1data) if rv1data is not None else rv1data
    rv2_smooth = smooth_rv(rv2data) if rv2data is not None else rv2data

    q, vgamma = estimate_q_vgamma(rv1_smooth, rv2_smooth)
    asinis = estimate_asini(rv1_smooth, rv2_smooth, period = 1.*u.d, vgamma = vgamma, ecc=0.)
    ph_supconj = estimate_phase_supconj(rv1_smooth, rv2_smooth, vgamma)
    # set initial values for ecc and per0
    ecc_inits = [0., 0.4]
    per0_inits = [0., np.pi/2, np.pi]

    loglikes = np.zeros((2,3))
    results = np.zeros((2,3,2))
    for i,ecc in enumerate(ecc_inits):
        for j,per0 in enumerate(per0_inits):
            init_params = [ecc, per0]
            for k in range(maxiter):
                result = least_squares(loglike, x0=init_params, ftol=1e-8, xtol=1e-8,
                    #bounds = ((times.min(), 0, 0., rvs.min()),(times.min()+period, 2*np.pi, 0.9, rvs.max())),
                    bounds = ((0.,0.), (0.9, 2*np.pi)),
                    kwargs={'rv1data':rv1data, 'rv2data':rv2data,
                            'asini': asinis,
                            'vgamma': vgamma, 'ph_supconj':ph_supconj})
                asinis = estimate_asini(rv1data, rv2data, period = 1.*u.d, vgamma = vgamma, ecc=result.x[0])
                if np.abs((result.x[0] - init_params[0])/result.x[0]) < 1e-2 and np.abs((result.x[1] - init_params[1])/result.x[1]) < 1e-2:
                    break
                init_params = result.x

            loglikes[i,j] = loglike(result.x, rv1data, rv2data, asinis, vgamma, ph_supconj)
            results[i,j] = result.x

    [ecc, per0] = results.reshape(6,2)[np.argmax(loglikes.reshape(6))]
    return {'q':q, 'asini': np.array(asinis),
            'vgamma':vgamma, 'ecc':ecc, 'per0':per0, 'ph_supconj': ph_supconj}
            # 'rv1_analytic': rv_model(rv1data[:,0], t0, period, result.x[0], result.x[1], asini*period, q, vgamma, component=1),
            # 'rv2_analytic': rv_model(rv2data[:,0], t0, period, result.x[0], result.x[1], asini*period, q, vgamma, component=2)}




from phoebe.parameters import *
from phoebe import conf

### NOTE: if creating new parameters, add to the _forbidden_labels list in parameters.py

def _comments_params(**kwargs):
    """
    """
    params = []

    params += [StringParameter(qualifier='comments', value=kwargs.get('comments', ''), description='User-provided comments for these solver-options.  Feel free to place any notes here - if not overridden, they will be copied to any resulting solutions.')]
    return params

def nelder_mead(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    scipy.optimize.minimize(method='nelder-mead') backend.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('optimizer.nelder_mead')
    b.run_solver(kind='nelder_mead')
    ```

    Parallelization support: nelder_mead does not support parallelization.  If
    within mpi, parallelization will be handled at the compute-level (per-time)
    for the <phoebe.parameters.compute.phoebe> backend.

    Arguments
    ----------
    * `compute` (string, optional): compute options to use for the forward
        model.
    * `expose_lnlikelihoods` (bool, optional, default=False): whether to expose
        the initial and final lnprobabilities in the solution (will result in 2
        additional forward model calls)
    * `fit_parameters` (list, optional, default=[]): parameters (as twigs) to
        optimize.
    * `initial_values` (dict, optional, default={}): twig-value pairs to
        (optionally) override the current values in the bundle.  Any items not
        in `fit_parameters` will be silently ignored.
    * `priors` (list, optional, default=[]): distribution(s) to use for priors
        (constrained and unconstrained parameters will be included, covariances
        will be respected except for distributions merge via `priors_combine`).
    * `priors_combine` (str, optional, default='and'): Method to use to combine
        multiple distributions from priors for the same parameter.
        first: ignore duplicate entries and take the first in the priors parameter.
        and: combine duplicate entries via AND logic, dropping covariances.
        or: combine duplicate entries via OR logic, dropping covariances.
    * `maxiter` (int, optional, default=1e6): passed directly to
        scipy.optimize.minimize.  Maximum allowed number of iterations.
    * `maxfev` (int, optional, default=1e6): passed directly to
        scipy.optimize.minimize.  Maximum allowed number of function evaluations
        (forward models).
    * `adaptive` (bool, optional, default=False): passed directly to
        scipy.optimize.minimize.  Adapt algorithm parameters to dimensionality
        of problem. Useful for high-dimensional minimization
    * `xatol` (float, optional, default=1e-4): passed directly to
        scipy.optimize.minimize.  Absolute error in xopt (input parameters)
        between iterations that is acceptable for convergence.
    * `fatol` (float, optional, default=1e-4): passed directly to
        scipy.optimize.minimize.  Absolute error in func(xopt)
        (lnlikelihood + lnp(priors)) between iterations that is acceptable for
        convergence.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [ChoiceParameter(qualifier='compute', value=kwargs.get('compute', 'None'), choices=['None'], description='compute options to use for forward model')]
    params += [BoolParameter(qualifier='expose_lnprobabilities', value=kwargs.get('expose_lnprobabilities', False), description='whether to expose the initial and final lnprobabilities in the solution (will result in 2 additional forward model calls)')]

    params += [SelectTwigParameter(qualifier='fit_parameters', value=kwargs.get('fit_parameters', []), choices=[], description='parameters (as twigs) to optimize')]
    params += [DictParameter(qualifier='initial_values', value=kwargs.get('initial_values', {}), description='twig-value pairs to (optionally) override the current values in the bundle.  Any items not in fit_parameters will be silently ignored.')]

    params += [SelectParameter(qualifier='priors', value=kwargs.get('priors', []), choices=[], description='distribution(s) to use for priors (constrained and unconstrained parameters will be included, covariances will be respected except for distributions merge via priors_combine)')]
    params += [ChoiceParameter(visible_if='priors:<notempty>', qualifier='priors_combine', value=kwargs.get('priors_combine', 'and'), choices=['first', 'and', 'or'], description='Method to use to combine multiple distributions from priors for the same parameter.  first: ignore duplicate entries and take the first in the priors parameter. and: combine duplicate entries via AND logic, dropping covariances.  or: combine duplicate entries via OR logic, dropping covariances.')]

    params += [IntParameter(qualifier='maxiter', value=kwargs.get('maxiter', 1e6), limits=[1,1e12], description='passed directly to scipy.optimize.minimize.  Maximum allowed number of iterations.')]
    params += [IntParameter(qualifier='maxfev', value=kwargs.get('maxfev', 1e6), limits=[1,1e12], description='passed directly to scipy.optimize.minimize.  Maximum allowed number of function evaluations (forward models).')]
    params += [BoolParameter(qualifier='adaptive', value=kwargs.get('adaptive', False), description='passed directly to scipy.optimize.minimize.  Adapt algorithm parameters to dimensionality of problem. Useful for high-dimensional minimization')]

    params += [FloatParameter(qualifier='xatol', value=kwargs.get('xatol', 1e-4), limits=[1e-12,None], description='passed directly to scipy.optimize.minimize.  Absolute error in xopt (input parameters) between iterations that is acceptable for convergence.')]
    params += [FloatParameter(qualifier='fatol', value=kwargs.get('fatol', 1e-4), limits=[1e-12,None], description='passed directly to scipy.optimize.minimize.  Absolute error in func(xopt) (lnlikelihood + lnp(priors)) between iterations that is acceptable for convergence.')]

    return ParameterSet(params)

def powell(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    scipy.optimize.minimize(method='powell') backend.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('optimizer.powell')
    b.run_solver(kind='powell')
    ```

    Parallelization support: powell does not support parallelization.  If
    within mpi, parallelization will be handled at the compute-level (per-time)
    for the <phoebe.parameters.compute.phoebe> backend.

    Arguments
    ----------
    * `compute` (string, optional): compute options to use for the forward
        model.
    * `expose_lnprobabilities` (bool, optional, default=False): whether to expose
        the initial and final lnlikelihoods in the solution (will result in 2
        additional forward model calls)
    * `fit_parameters` (list, optional, default=[]): parameters (as twigs) to
        optimize.
    * `initial_values` (dict, optional, default={}): twig-value pairs to
        (optionally) override the current values in the bundle.  Any items not
        in `fit_parameters` will be silently ignored.
    * `priors` (list, optional, default=[]): distribution(s) to use for priors
        (constrained and unconstrained parameters will be included, covariances
        will be respected except for distributions merge via `priors_combine`).
    * `priors_combine` (str, optional, default='and'): Method to use to combine
        multiple distributions from priors for the same parameter.
        first: ignore duplicate entries and take the first in the priors parameter.
        and: combine duplicate entries via AND logic, dropping covariances.
        or: combine duplicate entries via OR logic, dropping covariances.
    * `maxiter` (int, optional, default=1e6): passed directly to
        scipy.optimize.minimize.  Maximum allowed number of iterations.
    * `maxfev` (int, optional, default=1e6): passed directly to
        scipy.optimize.minimize.  Maximum allowed number of function evaluations
        (forward models).
    * `xtol` (float, optional, default=1e-4): passed directly to
        scipy.optimize.minimize.  Relative error in xopt (input parameters)
        between iterations that is acceptable for convergence.
    * `ftol` (float, optional, default=1e-4): passed directly to
        scipy.optimize.minimize.  Relative error in func(xopt)
        (lnlikelihood + lnp(priors)) between iterations that is acceptable for
        convergence.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [ChoiceParameter(qualifier='compute', value=kwargs.get('compute', 'None'), choices=['None'], description='compute options to use for forward model')]
    params += [BoolParameter(qualifier='expose_lnprobabilities', value=kwargs.get('expose_lnprobabilities', False), description='whether to expose the initial and final lnprobabilities in the solution (will result in 2 additional forward model calls)')]

    params += [SelectTwigParameter(qualifier='fit_parameters', value=kwargs.get('fit_parameters', []), choices=[], description='parameters to optimize')]
    params += [DictParameter(qualifier='initial_values', value=kwargs.get('initial_values', {}), description='twig-value pairs to (optionally) override the current values in the bundle.  Any items not in fit_parameters will be silently ignored.')]

    params += [SelectParameter(qualifier='priors', value=kwargs.get('priors', []), choices=[], description='distribution(s) to use for priors (constrained and unconstrained parameters will be included, covariances will be respected except for distributions merge via priors_combine)')]
    params += [ChoiceParameter(visible_if='priors:<notempty>', qualifier='priors_combine', value=kwargs.get('priors_combine', 'and'), choices=['first', 'and', 'or'], description='Method to use to combine multiple distributions from priors for the same parameter.  irst: ignore duplicate entries and take the first in the priors parameter. and: combine duplicate entries via AND logic, dropping covariances.  or: combine duplicate entries via OR logic, dropping covariances.')]

    params += [IntParameter(qualifier='maxiter', value=kwargs.get('maxiter', 1e6), limits=[1,1e12], description='passed directly to scipy.optimize.minimize.  Maximum allowed number of iterations.')]
    params += [IntParameter(qualifier='maxfev', value=kwargs.get('maxfev', 1e6), limits=[1,1e12], description='passed directly to scipy.optimize.minimize.  Maximum allowed number of function evaluations (forward models).')]

    params += [FloatParameter(qualifier='xtol', value=kwargs.get('xtol', 1e-4), limits=[1e-12,None], description='passed directly to scipy.optimize.minimize.  Relative error in xopt (input parameters) between iterations that is acceptable for convergence.')]
    params += [FloatParameter(qualifier='ftol', value=kwargs.get('ftol', 1e-4), limits=[1e-12,None], description='passed directly to scipy.optimize.minimize.  Relative error in func(xopt) (lnlikelihood + lnp(priors)) between iterations that is acceptable for convergence.')]

    return ParameterSet(params)

def cg(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    scipy.optimize.minimize(method='cg') "conjugate gradient" backend.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('optimizer.cg')
    b.run_solver(kind='cg')
    ```

    Parallelization support: cg does not support parallelization.  If
    within mpi, parallelization will be handled at the compute-level (per-time)
    for the <phoebe.parameters.compute.phoebe> backend.

    Arguments
    ----------
    * `compute` (string, optional): compute options to use for the forward
        model.
    * `expose_lnlikelihoods` (bool, optional, default=False): whether to expose
        the initial and final lnprobabilities in the solution (will result in 2
        additional forward model calls)
    * `fit_parameters` (list, optional, default=[]): parameters (as twigs) to
        optimize.
    * `initial_values` (dict, optional, default={}): twig-value pairs to
        (optionally) override the current values in the bundle.  Any items not
        in `fit_parameters` will be silently ignored.
    * `priors` (list, optional, default=[]): distribution(s) to use for priors
        (constrained and unconstrained parameters will be included, covariances
        will be respected except for distributions merge via `priors_combine`).
    * `priors_combine` (str, optional, default='and'): Method to use to combine
        multiple distributions from priors for the same parameter.
        first: ignore duplicate entries and take the first in the priors parameter.
        and: combine duplicate entries via AND logic, dropping covariances.
        or: combine duplicate entries via OR logic, dropping covariances.
    * `maxiter` (int, optional, default=1e6): passed directly to
        scipy.optimize.minimize.  Maximum allowed number of iterations.
    * `gtol` (float, optional, default=1e-5): passed directly to
        scipy.optimize.minimize.  Gradient norm must be less than gtol before successful termination.
    * `norm` (float, optional, default=np.inf): passed directly to
        scipy.optimize.minimize.  Order of norm (Inf is max, -Inf is min).

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _comments_params(**kwargs)

    params += [ChoiceParameter(qualifier='compute', value=kwargs.get('compute', 'None'), choices=['None'], description='compute options to use for forward model')]
    params += [BoolParameter(qualifier='expose_lnprobabilities', value=kwargs.get('expose_lnprobabilities', False), description='whether to expose the initial and final lnprobabilities in the solution (will result in 2 additional forward model calls)')]

    params += [SelectTwigParameter(qualifier='fit_parameters', value=kwargs.get('fit_parameters', []), choices=[], description='parameters to optimize')]
    params += [DictParameter(qualifier='initial_values', value=kwargs.get('initial_values', {}), description='twig-value pairs to (optionally) override the current values in the bundle.  Any items not in fit_parameters will be silently ignored.')]

    params += [SelectParameter(qualifier='priors', value=kwargs.get('priors', []), choices=[], description='distribution(s) to use for priors (constrained and unconstrained parameters will be included, covariances will be respected except for distributions merge via priors_combine)')]
    params += [ChoiceParameter(visible_if='priors:<notempty>', qualifier='priors_combine', value=kwargs.get('priors_combine', 'and'), choices=['first', 'and', 'or'], description='Method to use to combine multiple distributions from priors for the same parameter.  irst: ignore duplicate entries and take the first in the priors parameter. and: combine duplicate entries via AND logic, dropping covariances.  or: combine duplicate entries via OR logic, dropping covariances.')]

    params += [IntParameter(qualifier='maxiter', value=kwargs.get('maxiter', 1e6), limits=[1,1e12], description='passed directly to scipy.optimize.minimize.  Maximum allowed number of iterations.')]

    params += [FloatParameter(qualifier='gtol', value=kwargs.get('gtol', 1e-5), limits=[1e-12,None], description='passed directly to scipy.optimize.minimize.  Gradient norm must be less than gtol before successful termination.')]
    params += [FloatParameter(qualifier='norm', value=kwargs.get('norm', np.inf), limits=[None,None], description='passed directly to scipy.optimize.minimize.  Order of norm (Inf is max, -Inf is min).')]

    return ParameterSet(params)

def differential_evolution(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for solver options for the
    scipy.optimize.differential_evolution() backend.

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_solver>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_solver>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_solver('optimizer.differential_evolution')
    b.run_solver(kind='differential_evolution')
    ```

    Parallelization support: differential_evolution supports both MPI and multiprocessing, always
    at the solver-level (per-model).


    Arguments
    ----------

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    if not conf.devel:
        raise ImportError("differential_evolution is not fully tested and released.  Use developer mode to test.")

    params = _comments_params(**kwargs)

    params += [ChoiceParameter(qualifier='compute', value=kwargs.get('compute', 'None'), choices=['None'], description='compute options to use for forward model')]
    params += [BoolParameter(qualifier='expose_lnprobabilities', value=kwargs.get('expose_lnprobabilities', False), description='whether to expose the initial and final lnprobabilities in the solution (will result in 2 additional forward model calls)')]

    params += [SelectTwigParameter(qualifier='fit_parameters', value=kwargs.get('fit_parameters', []), choices=[], description='parameters to optimize')]

    params += [SelectParameter(qualifier='bounds', value=kwargs.get('bounds', []), choices=[], description='distribution(s) to use for bounds.  For any non-uniform distributions, bounds will be adopted based on the bounds_sigma parameter.  Only those in fit_parameters will be considered.  Any in fit_parameters but not in bounds will use the limits on the parameter itself as bounds.  Any covariances will be ignored.')]
    params += [ChoiceParameter(visible_if='bounds:<notempty>', qualifier='bounds_combine', value=kwargs.get('bounds_combine', 'first'), choices=['first', 'and', 'or'], description='Method to use to combine multiple distributions from bounds for the same parameter.  irst: ignore duplicate entries and take the first in the bounds parameter. and: combine duplicate entries via AND logic, dropping covariances.  or: combine duplicate entries via OR logic, dropping covariances..')]
    params += [FloatParameter(visible_if='bounds:<notempty>', qualifier='bounds_sigma', value=kwargs.get('bounds_sigma', 3), limits=(0,10), default_units=u.dimensionless_unscaled, description='sigma-level to use when converting non-uniform distributions for bounds to uniform bounds')]

    # params += [SelectParameter(qualifier='priors', value=kwargs.get('priors', []), choices=[], description='distribution(s) to use for priors')]
    # params += [ChoiceParameter(visible_if='priors:<notempty>', qualifier='priors_combine', value=kwargs.get('priors_combine', 'and'), choices=['first', 'and', 'or'], description='Method to use to combine multiple distributions from priors for the same parameter.  irst: ignore duplicate entries and take the first in the priors parameter. and: combine duplicate entries via AND logic, dropping covariances.  or: combine duplicate entries via OR logic, dropping covariances.')]

    strategy_choices = ['best1bin', 'best1exp', 'rand1exp', 'randtobest1exp', 'currenttobest1exp',
                        'best2exp', 'rand2exp', 'randtobest1bin', 'currenttobest1bin', 'best2bin',
                        'rand2bin', 'rand1bin']
    params += [ChoiceParameter(qualifier='strategy', value=kwargs.get('strategy', 'best1bin'), choices=strategy_choices, description='passed directly to scipy.optimize.differential_evolution.')]


    params += [IntParameter(qualifier='maxiter', value=kwargs.get('maxiter', 1e6), limits=[1,1e12], description='passed directly to scipy.optimize.differential_evolution.  The maximum number of generations over which the entire population is evolved. The maximum number of function evaluations (with no polishing) is: (maxiter + 1) * popsize * len(x)')]
    params += [IntParameter(qualifier='popsize', value=kwargs.get('popsize', 8), limits=[1,1e4], description='passed directly to scipy.optimize.differential_evolution.  A multiplier for setting the total population size. The population has popsize * len(x) individuals (unless the initial population is supplied via the init keyword)')]

    # TODO: expose mutation, recombination, seed, tol, atol, polish, init
    return ParameterSet(params)
